@@ -1,8 +1,8 @@
-/* Time-stamp: <2006-07-05 17:38:59 poser>
+/* Time-stamp: <2007-03-02 18:25:02 poser>
  *
  * Convert text containing various 7-bit ASCII escapes to UTF-7 Unicode.
  *
- * Copyright (C) 2005-2006 William J. Poser (billposer@alum.mit.edu)
+ * Copyright (C) 2005-2007 William J. Poser (billposer@alum.mit.edu)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -38,6 +38,8 @@
 #include "unicode.h"
 #include "enttbl.h"
 #include "exitcode.h"
+#include "formats.h"
+
 #if defined(__DATE__) && defined(__TIME__)
 #define HAVE_DATE_TIME
 char compdate[]= "Compiled " __DATE__ " " __TIME__ ;
@@ -48,7 +50,11 @@ char compdate[]= "";
 char version[]=PACKAGE_VERSION;
 char pgname[]="ascii2uni";
 
-#define LBUFSIZE 2048
+#ifndef LOCALEDIR 
+#define LOCALEDIR "/usr/local/share/locale"
+#endif
+
+#define LBUFSIZE 10240
 
 void
 ShowVersion(void){
@@ -56,8 +62,12 @@ ShowVersion(void){
 #ifdef HAVE_DATE_TIME
   fprintf(stderr,"%s\n",compdate);
 #endif
-  fprintf(stderr,"Copyright (C) 2005-2006 William J. Poser\n");
-  fprintf(stderr,_("Released under the terms of the GNU General Public License.\n\n"));
+  fprintf(stderr,"Copyright (C) 2004-2007 William J. Poser\n");
+  fprintf(stderr,"This program is free software; you can redistribute\n\
+it and/or modify it under the terms of version 2 of\n\
+the GNU General Public License as published by the\n\
+Free Software Foundation.\n");
+  fprintf(stderr,"Report bugs to: billposer@alum.mit.edu.\n");
 }
 
 void
@@ -65,98 +75,104 @@ ShowUsage(void){
   fprintf(stderr,_("This program is a filter which converts 7-bit ASCII text\n\
 containing various representations for non-ASCII characters\nto UTF-8 Unicode.\n"));
   fprintf(stderr,_("Usage: %s [flags]\n"),pgname);
+  fprintf(stderr,_("       -a <format specification>.\n"));
   fprintf(stderr,_("       -h Print this usage message.\n"));
+  fprintf(stderr,_("       -L List format codes.\n"));
+  fprintf(stderr,_("       -p Input consists of pure escapes except for non-null whitespace.\n"));
   fprintf(stderr,_("       -q Quiet - don't chat.\n"));
   fprintf(stderr,_("       -v Print version information.\n"));
   fprintf(stderr,_("       -8 Convert only tokens above the ASCII range.\n"));
-  fprintf(stderr,_("   Give at most one of the following conversion specifications:\n"));
-  fprintf(stderr,
-	  _("       -A Convert hexadecimal numbers with prefix U in angle-brackets(<U00E9>)\n"));
-  fprintf(stderr,
-	  _("       -B Convert \\x-escaped hexadecimal numbers (\\x00E9)\n"));
-  fprintf(stderr,
-	  _("       -C Convert \\x-escaped hexadecimal numbers in braces (\\x{00E9})\n"));
-  fprintf(stderr,
-	  _("       -D Convert decimal HTML numeric character references (&#0233;)\n"));
-  fprintf(stderr,
-	  _("       -E Convert hexadecimal with prefix U (U00E9)\n"));
-  fprintf(stderr,
-	  _("       -F Convert hexadecimal with prefix u (u00E9)\n"));
-  fprintf(stderr,
-	  _("       -G Convert hexadecimal in single quotes with prefix X (X\'00E9\')\n"));
-  fprintf(stderr,
-	  _("       -H Convert hexadecimal HTML numeric character references (&#x00E9;)\n"));
-  fprintf(stderr,
-	  _("       -I Convert hexadecimal UTF-8 with each byte's hex preceded by an =-sign (=C3=A9)\n\t\tThis is the URI escape format defined by RFC 2396.\n"));
-  fprintf(stderr,
-	  _("       -J Convert hexadecimal UTF-8 with each byte's hex preceded by a %%-sign  (%%C3%%A9).\n\t\tThis is the Quoted Printable format defined by RFC 2045.\n"));
-  fprintf(stderr,
-	  _("       -K Convert octal UTF-8 with each byte escaped by a backslash (\\303\\251)\n"));
-  fprintf(stderr,
-	  _("       -L Convert \\u-escaped hex (\\u00E9) within the BMP (U+0000-U+FFFF),\n\t\t\\U-escaped hex (\\U00010024) outisde it.\n"));
-  fprintf(stderr,
-	  _("       -M Convert hexadecimal SGML numeric character references (\\#x00E9;)\n"));
-  fprintf(stderr,
-	  _("       -N Convert decimal SGML numeric character references (\\#0233;)\n"));
-  fprintf(stderr,
-	  _("       -O Convert octal escapes for the three low bytes in big-endian order (\\000\\000\\351)\n"));
-  fprintf(stderr,
-	  _("       -P Convert hexadecimal numbers with prefix U+ (U+00E9)\n"));
-  fprintf(stderr,
-	  _("       -Q Convert HTML character entities (&eacute;)\n"));
-  fprintf(stderr,
-	  _("       -R Convert raw hexadecimal numbers (00E9)\n"));
-  fprintf(stderr,
-	  _("       -S Convert hexadecimal escapes for the three low bytes in big-endian order (\\x00\\x00\\xE9)\n"));
-  fprintf(stderr,
-	  _("       -T Convert decimal escapes for the three low bytes in big-endian order (\\d000\\d000\\d233)\n"));
-  fprintf(stderr,
-	  _("       -U Convert \\u-escaped hex (\\u00E9)\n"));
-  fprintf(stderr,
-	  _("       -V Convert \\u-escaped decimal (\\u0233)\n"));
-  fprintf(stderr,
-	  _("       -X Convert standard form hexadecimal numbers (0x00E9)\n"));
-  fprintf(stderr,
-	  _("       -Y Convert all three HTML escape types:\n\t\thexadecimal numeric, decimal numeric, and character entity.\n"));
   fprintf(stderr, 
 	  _("       -Z <format> Convert input using the supplied format.\n"));
   fprintf(stderr,_("Report bugs to: billposer@alum.mit.edu\n"));
 }
 
-static char lbuf [LBUFSIZE+1];
+
+char *
+ExtractSubstring(char *strptr, char* Start, char* End) {
+  char *i;
+  char *SavedBeginning;
+
+  SavedBeginning = strptr;
+  for (i = Start; i <= End; i++) *strptr++ = *i;
+  *strptr = '\0';
+  return SavedBeginning;
+}
 
 
 /* The length of the longest character entity */
 #define MAXENTLEN 8
 
+/* The library function seems not to be working. Anyhow, we want to keep this
+ * independent of locale.
+ */
+int myisxdigit (int c) {
+  switch (c) {
+  case '0':
+  case '1':
+  case '2':
+  case '3':
+  case '4':
+  case '5':
+  case '6':
+  case '7':
+  case '8':
+  case '9':
+  case 'a':
+  case 'b':
+  case 'c':
+  case 'd':
+  case 'e':
+  case 'f':
+  case 'A':
+  case 'B':
+  case 'C':
+  case 'D':
+  case 'E':
+  case 'F':
+    return 1;
+  default:
+    return 0;
+  }
+}
+
+static char *Formats [] = {
+"&#x%lX;", 	/* HTMLX */
+"&#%ld;", 	/* HTMLD */
+"\\#x%lX;", 	/* SGMLX */
+"\\#%ld;",	/* SGMLD */
+"\\u%8lX",	/* BSLU */
+"\\x%lX",	/* BSLX */
+"0x%4lX",	/* STDX */
+"#x%4lX",	/* CLSX */
+"%lX", 		/* RAWX */
+"\\x{%lX}",	/* BSLXB */
+"<U%lX>",	/* ABUX */
+"U%lX",		/* JUUX */
+"u%lX",		/* JuUX */
+"U+%lX",	/* UPLX */
+"X\'%lX\'",	/* XQ */
+"\\u%8ld",	/* BSLUD */
+"v%ld",		/* PERLV */
+"$%04X",	/* DOLLAR */
+"16#%04X",	/* PSPT */
+"#16r%04X",	/* CLR */
+"16#%04X#",	/* ADA */
+"\\%03o\\%03o\\%03o",	/* BYTEO */
+"\\d%03d\\d%03d\\d%03d",	/* BYTED */
+"\\x%02x\\x%02x\\x%02x",	/* BYTEX */
+"&%[abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789];", /* CHENT */
+"=%2lX", 		/* UTF-8  - Ifmt */
+"%%%2lX", 		/* UTF-8  - Jfmt */
+"\\%3lo" 		/* UTF-8  - Kfmt */
+};
+
+
 int main (int ac, char *av[])
 {
   char *SplitFormat = "\\%1[uU]%X%n"; /* This is for BMPSplit */
 
-  char *Afmt = "<U%lX>";
-  char *Bfmt = "\\x%lX";
-  char *Cfmt = "\\x{%lX}";
-  char *Dfmt = "&#%ld;"; 
-  char *Efmt = "U%lX";
-  char *Ffmt = "u%lX";
-  char *Gfmt = "X\'%lX\'";
-  char *Hfmt = "&#x%lX;"; 
-  char *Ifmt = "=%2lX"; 		/* UTF-8 */
-  char *Jfmt = "%%%2lX"; 		/* UTF-8 */
-  char *Kfmt = "\\%3lo"; 		/* UTF-8 */
-  char *Mfmt = "\\#x%lX;"; 
-  char *Nfmt = "\\#%ld;";
-  char *Ofmt = "\\%03o\\%03o\\%03o";
-  char *Pfmt = "U+%lX";
-  char *Qfmt = "&%[abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789];";
-  char *Rfmt = "%lX"; 
-  char *Sfmt = "\\x%02x\\x%02x\\x%02x";	
-  char *Tfmt = "\\d%03d\\d%03d\\d%03d";
-  char *Ufmt = "\\u%8lX";
-  char *Vfmt = "\\u%8ld";
-  char *Xfmt = "0x%4lX";
-
-  char *fmt = Xfmt;		/* Default is plain hex format */
+  char *fmt = Formats[STDX];		/* Default is plain hex format */
   char afmt [67+2+1+2];
   char aHfmt [8+2+1];
   char aDfmt [8+2+1];
@@ -167,10 +183,10 @@ int main (int ac, char *av[])
   int Converted;
   long TokenNumber;
   long ReplacedNumber;
-  int BMPSplit = 0; 
+  short BMPSplitP = 0; 
   int VerboseP = 1;
   int UTF8ValueP = 0;		/* Are incoming values UTF-8? */
-  int AllHTMLP = 0; 		/* Translate all three kinds of HTML escape */
+  short AllHTMLP = 0; 		/* Translate all three kinds of HTML escape */
   int PureP = 0;
   int Word_Length;
   int NConsumed;
@@ -181,9 +197,16 @@ int main (int ac, char *av[])
   char SplitStr[3];
   char enam[LBUFSIZE];
   long MicrosoftStyle = 0L;
+  char tmpstr [16];
   unsigned char b1;		/* Used for byte-wise encoding */
   unsigned char b2;
   unsigned char b3;
+  int FType;
+  int UTF8Type;			/* Not used - for compatibility with uni2ascii */
+  short UseEntitiesP;	/* Not used - for compatibility with uni2ascii */
+  int last;
+
+  char lbuf [LBUFSIZE+1];
 
   extern int optind;
   extern int opterr;
@@ -191,6 +214,10 @@ int main (int ac, char *av[])
   extern void putu8 (unsigned long);
   extern char * Get_Word(FILE *, int *, int *);
 
+  extern int CountSlots(char *);
+  extern void ListFormatArguments(short);
+  extern void SetFormat(char *, int *, short *,int *, short *, short *);
+
   opterr = 0;
 
 #ifdef HAVE_SETLOCALE
@@ -203,90 +230,28 @@ int main (int ac, char *av[])
 
   /* Handle command line arguments */
 
-   while( (oc = getopt(ac,av,"ABCDEFHIJKLMNOPQRSTUVXYZ:hpqv")) != EOF){
+   while( (oc = getopt(ac,av,":Z:a:hLpqv")) != EOF){
      switch(oc){
-     case 'A':
-       fmt = Afmt;
-       break;
-     case 'X':
-       fmt = Xfmt;
-       break;
-     case 'O':
-       fmt = Ofmt;
-       break;
-     case 'S':
-       fmt = Sfmt;
-       break;
-     case 'T':
-       fmt = Tfmt;
-       break;
-     case 'P':
-       fmt = Pfmt;
-       break;
-     case 'Q':
-       fmt = Qfmt;
-       break;
-     case 'R':
-       fmt = Rfmt;
-       break;
-     case 'B':
-       fmt = Bfmt;
-       break;
-     case 'C':
-       fmt = Cfmt;
-       break;
-     case 'D':
-       fmt = Dfmt;
-       break;
-     case 'E':
-       fmt = Efmt;
-       break;
-     case 'F':
-       fmt = Ffmt;
-       break;
-     case 'G':
-       fmt = Gfmt;
-       break;
-     case 'H':
-       fmt = Hfmt;
-       break;
-     case 'I':
-       fmt = Ifmt;
-       UTF8ValueP = 1;
-       break;
-     case 'J':
-       fmt = Jfmt;
-       UTF8ValueP = 1;
-       cbuf[0] = '0';
-       cbuf[1] = 'x';
-       break;
-     case 'K':
-       fmt = Kfmt;
-       cbuf[0] = '\\';
-       UTF8ValueP = 1;
-       break;
-     case 'L':
-       fmt = SplitFormat;
-       BMPSplit =1;
-       break;
-     case 'M':
-       fmt = Mfmt;
-       break;
-     case 'N':
-       fmt = Nfmt;
-       break;
-     case 'U':
-       fmt = Ufmt;
-       break;
-     case 'V':
-       fmt = Vfmt;
-       break;
-     case 'Y':
-       fmt = Qfmt;
-       AllHTMLP = 1;
+     case 'a':
+       SetFormat(optarg,&FType,&UseEntitiesP, &UTF8Type, &BMPSplitP,&AllHTMLP);
+       if(FType == FMT_UNKNOWN) {
+	 fprintf(stderr,"Format specification %s not recognized.\n",optarg);
+	 exit(BADOPTIONARG);
+       }
+       fmt = Formats[FType];
+       if((FType == IFMT) || (FType == JFMT) || (FType == KFMT)) UTF8ValueP =1;
+       if(FType == JFMT) {cbuf[0] = '0'; cbuf[1] = 'x';}
+       if(FType == KFMT) {cbuf[0] = '\\';}
        break;
+    case 'L':
+      ListFormatArguments(0);
+      exit(INFO);
      case 'Z':
        fmt = optarg;
+       if(CountSlots(fmt) > 1) {
+	 fprintf(stderr,"You may not supply a format with more than one empty slot.\n");
+	 exit(BADOPTIONARG);
+       }
        break;
      case 'p':
        PureP = 1;
@@ -313,8 +278,8 @@ int main (int ac, char *av[])
      }
    }
 
-   if( (fmt == Rfmt) && (!PureP) ) {
-     fprintf(stderr,_("It isn't possible to parse raw hex unicode out of ASCII text.\n"));
+   if( (FType == RAWX) && (!PureP) ) {
+     fprintf(stderr,_("It isn't possible reliably to parse raw hex unicode out of ASCII text.\n"));
      exit(BADOPTION);
    }
 
@@ -324,13 +289,14 @@ int main (int ac, char *av[])
    }
 
    if(AllHTMLP) {
-     sprintf(aDfmt,"%s%%n",Dfmt);
-     sprintf(aHfmt,"%s%%n",Hfmt);
+     sprintf(aDfmt,"%s%%n",Formats[HTMLD]);
+     sprintf(aHfmt,"%s%%n",Formats[HTMLX]);
    }
 
    sprintf(afmt,"%s%%n",fmt);	/* Add %n for NConsumed */
    ReplacedNumber = 0L;
    TokenNumber = 0L;
+
    /*
     * This is the case in which the input consists entirely of escapes
     * except for arbitrary (but non-null) amounts of intervening whitespace.
@@ -347,7 +313,7 @@ int main (int ac, char *av[])
 		 pgname,TokenNumber);
 	 exit(OUTOFMEMORY);
        }
-       if(fmt == Qfmt) {
+       if(FType == CHENT) {
 	 Converted = sscanf(str,afmt,&enam,&NConsumed);
 	 num = LookupCodeForEntity(enam);
 	 if(!num) {
@@ -359,7 +325,7 @@ int main (int ac, char *av[])
 	 }
 	 else Converted = 1;
        }
-       else if( (Ofmt == fmt) || (Sfmt == fmt) || (Tfmt == fmt)) {
+       else if( (BYTEO == FType) || (BYTED == FType) || (BYTEH == FType)) {
 	 Converted = sscanf(str,afmt,&b1,&b2,&b3,&NConsumed);
 	 switch(Converted)
 	   {
@@ -393,33 +359,38 @@ int main (int ac, char *av[])
        else {
 	 if (UTF8ValueP) putchar(num);
 	 else putu8(num);
-	 if( (fmt == Dfmt) || (fmt == Hfmt) || (fmt == Qfmt)) {
-	   if(*(str+NConsumed-1) != ';') MicrosoftStyle++;
+	 if( (FType == HTMLD) || (FType == HTMLX) || (FType == CHENT)) {
+	   if(*(str+NConsumed-1) != ';') {
+	     MicrosoftStyle++;
+	     fprintf(stderr,_("The HTML entity %1$s at token %2$lu lacks the requisite final semicolon.\n"),str,TokenNumber);
+	   }
 	 }
        }
        free((void *)str);
      }
      goto done;
-   }
+   } /* End of PureP */
 
    /* This is the case in which the Unicode escapes are embedded in ASCII text */
 
    LineNo = 0;
    while(fgets(lbuf,LBUFSIZE,stdin) != NULL) {
      LineNo++;
+     last = strlen(lbuf) - 1;
+     if(lbuf[last] == '\n') {lbuf[last] = '\0'; last--;}
+     if(last < 0) continue;
      iptr = lbuf;
-     if(fmt == Jfmt) {
+     if(FType == JFMT) {
        while(*iptr) {
 	 if(*iptr == '%') {
 	   if(*++iptr) {
-	     if(isxdigit(*iptr++)) {
-	       if(*iptr) {
-		 if(isxdigit(*iptr)) { /* match */
+	     if(myisxdigit(*iptr)) {
+	       if(*++iptr) {
+		 if(myisxdigit(*iptr)) { /* match */
 		   cbuf[2] = *(iptr-1);
 		   cbuf[3] = *iptr;
 		   cbuf[4] = '\0';
-   /*		   fprintf(stderr,"cbuf = %s\n",cbuf); */
-		   num = strtoul(cbuf,NULL,16);
+		   num = (unsigned char)strtoul(cbuf,NULL,16);
 		   putchar(num);
 		   TokenNumber++;
 		   iptr++;
@@ -457,8 +428,8 @@ int main (int ac, char *av[])
        }
      } /* End of special case for J format */
 
-     while (*iptr) { 
-       if(BMPSplit) {
+     while (*iptr) {
+       if(BMPSplitP) {
 	 if(sscanf(iptr,SplitFormat,&SplitStr,&num,&NConsumed)) {
 	   if( (num <= 0xFFFF) && (SplitStr[0] == 'U')) {
 	     fprintf(stderr,_("Warning: the code \\U%1$08lX at line %2$d falls within the BMP.\n"),
@@ -474,19 +445,25 @@ int main (int ac, char *av[])
 	 }
 	 else putchar(*iptr++);
        }
-       else if (fmt == Qfmt) {
+       else if (FType == CHENT) {
 	 if (AllHTMLP){
 	   if(sscanf(iptr,aHfmt,&num,&NConsumed)) {
 	     putu8(num);
 	     iptr+=NConsumed;
-	     if(*(iptr-1) != ';') MicrosoftStyle++;
+	     if(*(iptr-1) != ';') {
+	       MicrosoftStyle++;
+	       fprintf(stderr,_("The HTML entity %1$s at token %2$lu lacks the requisite final semicolon.\n"),ExtractSubstring(tmpstr,iptr-NConsumed,iptr-1),TokenNumber);
+	     }
 	     TokenNumber++;
 	     continue;
 	   }
 	   else if(sscanf(iptr,aDfmt,&num,&NConsumed)) {
 	     putu8(num);
 	     iptr+=NConsumed;
-	     if(*(iptr-1) != ';') MicrosoftStyle++;
+	     if(*(iptr-1) != ';') {
+	       MicrosoftStyle++;
+	       fprintf(stderr,_("The HTML entity %1$s at token %2$lu lacks the requisite final semicolon.\n"),ExtractSubstring(tmpstr,iptr-NConsumed,iptr-1),TokenNumber);
+	     }
 	     TokenNumber++;
 	     continue;
 	   }
@@ -495,7 +472,10 @@ int main (int ac, char *av[])
 	   if( (num = LookupCodeForEntity(enam))) {
 	     putu8(num);
 	     iptr+=NConsumed;
-	     if(*(iptr-1) != ';') MicrosoftStyle++;
+	     if(*(iptr-1) != ';') {
+	       MicrosoftStyle++;
+	       fprintf(stderr,_("The HTML entity %1$s at token %2$lu lacks the requisite final semicolon.\n"),ExtractSubstring(tmpstr,iptr-NConsumed,iptr-1),TokenNumber);
+	     }
 	     TokenNumber++;
 	   }
 	   else {
@@ -508,7 +488,7 @@ int main (int ac, char *av[])
 	 }
 	 else putchar(*iptr++);
        } /* End of Qfmt case */
-       else if( (Ofmt == fmt) || (Sfmt == fmt) || (Tfmt == fmt)) {
+       else if( (BYTEO == FType) || (BYTED == FType) || (BYTEH == FType)) {
 	 Converted=sscanf(iptr,afmt,&b1,&b2,&b3,&NConsumed);
 	 switch(Converted)
 	   {
@@ -534,28 +514,42 @@ int main (int ac, char *av[])
 	   }
 	   TokenNumber++;
        }
-       else {			/* Default - not BMPSplit, HTML, or byte format */
-	  if(sscanf(iptr,afmt,&num,&NConsumed)) {
+       else {			/* Default - not BMPSplitP, Q, or byte format */
+	 //	 NConsumed=0;
+	 if(sscanf(iptr,afmt,&num,&NConsumed)) {
 	   if (UTF8ValueP) putchar(num);
 	   else putu8(num);
 	   iptr+=NConsumed;
-	   if(fmt == Hfmt) {
-	     if(*(iptr-1) != ';') MicrosoftStyle++;
+	   if(FType== HTMLX) {
+	     if(*(iptr-1) != ';') {
+	       MicrosoftStyle++;
+	       /*	       fprintf(stderr,_("The HTML entity %1$s at token %2$lu lacks the requisite final semicolon.\n"),ExtractSubstring(tmpstr,iptr-NConsumed,iptr-1),TokenNumber); */
+	       fprintf(stderr,"The HTML entity %s at token %lu lacks the requisite final semicolon.\n",ExtractSubstring(tmpstr,iptr-NConsumed,iptr-1),TokenNumber);
+	     }
 	   }
-	   else if(fmt == Dfmt) {
-	     if(*(iptr-1) != ';') MicrosoftStyle++;
+	   else if(FType == HTMLD) {
+	     if(*(iptr-1) != ';') {
+	       MicrosoftStyle++;
+	       fprintf(stderr,_("The HTML entity %1$s at token %2$lu lacks the requisite final semicolon.\n"),ExtractSubstring(tmpstr,iptr-NConsumed,iptr-1),TokenNumber);
+	     }
 	   }
 	   TokenNumber++;
 	 }
 	 else putchar(*iptr++);
        }
      } /* Loop over current line */
+     putchar('\n');
    } /* Loop over input lines */
 
 done:
    if(VerboseP) {
-     fprintf(stderr,_("%ld tokens converted\n"),TokenNumber);
-     fprintf(stderr,_("%ld tokens replaced with Unicode Replacement Character\n"),ReplacedNumber);
+     if (TokenNumber == 1)  fprintf(stderr,_("%ld token converted\n"),TokenNumber);
+     else fprintf(stderr,_("%ld tokens converted\n"),TokenNumber);
+     if (ReplacedNumber) {
+       if (ReplacedNumber == 1) fprintf(stderr,
+		_("%ld token replaced with Unicode Replacement Character\n"),ReplacedNumber);
+       else fprintf(stderr,_("%ld tokens replaced with Unicode Replacement Character\n"),ReplacedNumber);
+     }
      if(MicrosoftStyle) {
        fprintf(stderr,
 	       _("%ld Microsoft-style (lacking final semi-colon)\n"),MicrosoftStyle);
