@@ -2,9 +2,9 @@
 #-*- mode: Tcl;time-stamp-start:"TimeStamp[ 	]+\\\\?[\"<]+";-*-
 # the next line restarts using wish \
 exec wish $0 -- $@
-set TimeStamp "2006-07-03 22:14:03 poser"
+set TimeStamp "2007-03-11 20:52:56 poser"
 #
-# Copyright (C) 2005, 2006 William J. Poser (billposer@alum.mit.edu)
+# Copyright (C) 2005-2007 William J. Poser (billposer@alum.mit.edu)
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of version 2 of the GNU General Public License as
 # published by the Free Software Foundation.
@@ -20,7 +20,7 @@ set TimeStamp "2006-07-03 22:14:03 poser"
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 # or go to the web page:  http://www.gnu.org/licenses/gpl.txt.
 
-set Version "3.9.4";
+set Version "4.3"
 encoding system utf-8
 set DebugP 0;				# General debugging?
 package require tablelist
@@ -38,11 +38,12 @@ set ColorSpecs(Button,Background)		"\#8BD664"
 set ColorSpecs(Button,Foreground)		black
 set ColorSpecs(Button,ActiveBackground)		LightBlue
 set ColorSpecs(Button,ActiveForeground)		red
+set ColorSpecs(CheckButton,Background)		coral
 set ColorSpecs(CheckButton,ActiveForeground)	red
 set ColorSpecs(CheckButton,ActiveBackground)	LightBlue
 set ColorSpecs(CheckButton,Select)	red
 #set ColorSpecs(Default,Background)		"\#DEDEFF"
-set ColorSpecs(Default,Background)		"\#aff3b1"
+set ColorSpecs(Default,Background)		"\#c8dcfe"
 set ColorSpecs(DefaultActive,Background)	"\#d4e4fe";
 set ColorSpecs(Entry,Background)	"navajo white"
 set ColorSpecs(Entry,DisabledBackground)		blue
@@ -87,7 +88,7 @@ option add *Radiobutton.activeBackground  $ColorSpecs(RadioButton,ActiveBackgrou
 option add *Radiobutton.activeForeground  $ColorSpecs(RadioButton,ActiveForeground)
 option add *Radiobutton.selectColor  $ColorSpecs(RadioButton,Select) 80
 
-#option add *Checkbutton*background lightblue
+option add *Checkbutton*background $ColorSpecs(CheckButton,Background)
 option add *Checkbutton.activeBackground  $ColorSpecs(CheckButton,ActiveBackground)
 option add *Checkbutton.activeForeground  $ColorSpecs(CheckButton,ActiveForeground)
 option add *Checkbutton.selectColor  $ColorSpecs(CheckButton,Select)
@@ -199,6 +200,9 @@ proc DetermineGraphicsSystem {} {
 DetermineGraphicsSystem
 SetupEvents $SystemInfo(System)
 
+set SubstitutionList {}
+set DeletionList {}
+
 #Determines which program to use
 # 1 = ascii2uni
 # 0 = uni2ascii
@@ -435,14 +439,27 @@ proc ExplainExpansions {} {
     set po [CreateTextDisplay [_ "Expansions"] 72 12]
     set ::HelpPopups(ExplainExpansions) $po
     AppendToTextDisplay $po "If this option is chosen, certain characters are expanded to sequences of approximately equivalent plain ASCII characters. The expansions are:\n"
+    AppendToTextDisplay $po "\tU+00A2 \u00A2\tCENT SIGN         \t\u2192 cent\n"
+    AppendToTextDisplay $po "\tU+00A3 \u00A3\tPOUND SIGN        \t\u2192 pound\n"
+    AppendToTextDisplay $po "\tU+00A5 \u00A5\tYEN SIGN          \t\u2192 yen\n"
     AppendToTextDisplay $po "\tU+00A9 \u00A9\tCOPYRIGHT SYMBOL         \t\u2192 (c)\n"
-    AppendToTextDisplay $po "\tU+00DF \u00DF\tSMALL LETTER SHARP S     \t\u2192 ss\n"
+    AppendToTextDisplay $po "\tU+00AE \u00AE\tREGISTERED SYMBOL        \t\u2192 (R)\n"
+    AppendToTextDisplay $po "\tU+00BC \u00BC\tONE QUARTER              \t\u2192 1/4\n"
+    AppendToTextDisplay $po "\tU+00BD \u00BD\tONE HALF                 \t\u2192 1/2\n"
+    AppendToTextDisplay $po "\tU+00BE \u00BE\tTHREE QUARTERS           \t\u2192 3/4\n"
     AppendToTextDisplay $po "\tU+00C6 \u00C6\tCAPITAL LETTER ASH       \t\u2192 AE\n"
+    AppendToTextDisplay $po "\tU+00DF \u00DF\tSMALL LETTER SHARP S     \t\u2192 ss\n"
     AppendToTextDisplay $po "\tU+00E6 \u00E6\tSMALL LETTER ASH         \t\u2192 ae\n"
+    AppendToTextDisplay $po "\tU+0132 \u0132\tLIGATURE IJ              \t\u2192 IJ\n"
+    AppendToTextDisplay $po "\tU+0133 \u0133\tLIGATURE ij              \t\u2192 ij\n"
+    AppendToTextDisplay $po "\tU+0152 \u0152\tLIGATURE OE              \t\u2192 OE\n"
+    AppendToTextDisplay $po "\tU+0153 \u0153\tLIGATURE oe              \t\u2192 oe\n"
     AppendToTextDisplay $po "\tU+01F1 \u01F1\tCAPITAL LETTER DZ        \t\u2192 DZ\n"
+    AppendToTextDisplay $po "\tU+01F2 \u01F2\tCAPITAL LETTER Dz        \t\u2192 Dz\n"
     AppendToTextDisplay $po "\tU+01F3 \u01F3\tSMALLL LETTER DZ         \t\u2192 dz\n"
     AppendToTextDisplay $po "\tU+02A6 \u02A6\tSMALLL LETTER TS DIGRAPH \t\u2192 ts\n"
     AppendToTextDisplay $po "\tU+2026 \u2026\tHORIZONTAL ELLIPSIS      \t\u2192 ...\n"
+    AppendToTextDisplay $po "\tU+20AC \u20AC\tEURO SIGN                \t\u2192 euro\n"
     AppendToTextDisplay $po "\tU+2190 \u2190\tLEFTWARDS ARROW          \t\u2192 <-\n"
     AppendToTextDisplay $po "\tU+2192 \u2192\tRIGHTWARDS ARROW         \t\u2192 ->\n"
     AppendToTextDisplay $po "\tU+21D0 \u21D0\tLEFTWARDS DOUBLE ARROW   \t\u2192 <=\n"
@@ -460,7 +477,7 @@ proc About {} {
     AppendToTextDisplay $AboutPopup [_ "You can obtain the latest version of U2A from: "];
     AppendLinkToTextDisplay $AboutPopup "http://billposer.org/Software/uni2ascii.html." {ShowWebPage http://billposer.org/Software/uni2ascii.html};
     AppendToTextDisplay $AboutPopup "\n\n";
-    AppendToTextDisplay $AboutPopup "Copyright (C) 2005, 2006 William J. Poser (billposer@alum.mit.edu). ";
+    AppendToTextDisplay $AboutPopup "Copyright (C) 2005-2007 William J. Poser (billposer@alum.mit.edu). ";
     AppendToTextDisplay $AboutPopup [_ "This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version."];
     AppendToTextDisplay $AboutPopup "\n\n";
     AppendToTextDisplay $AboutPopup [_ "This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"];
@@ -803,7 +820,16 @@ proc HowTo {} {
 
 #The order in which the options are listed here is the order in which they will be
 #displayed.
-set OptionList {r x f e u v l a p b c g d h n m q y k i j o s t}
+set OptionList {r x f e u v l a p b c g 1 3 d 2 4 5 6 h n m q y k i j o s t}
+set AsciiOptions(1,example) "\#x00E9"
+set AsciiOptions(1,description) "Hexadecimal with \#x preceding four digits"
+set AsciiOptions(1,info) "This notation is used in Common Lisp."
+set AsciiOptions(2,example) "v233"
+set AsciiOptions(2,description) "Decimal with v preceding number"
+set AsciiOptions(2,info) "This notation is used in Perl."
+set AsciiOptions(3,example) "\$00E9"
+set AsciiOptions(3,description) "Hexadecimal with \$ preceding number"
+set AsciiOptions(3,info) "This notation is used in Pascal and by many assemblers."
 set AsciiOptions(a,example) "<U00E9>"
 set AsciiOptions(a,description) "hexadecimal with prefix U in angle brackets"
 set AsciiOptions(a,info) "This notation is used in POSIX locale specifications."
@@ -868,6 +894,15 @@ set AsciiOptions(x,example) "0x00E9"
 set AsciiOptions(x,description) "standard format hexadecimal number"
 set AsciiOptions(y,example) ""
 set AsciiOptions(y,description) "all three HTML formats"
+set AsciiOptions(4,description) "Generate hexadecimal numbers with prefix 16\#"
+set AsciiOptions(4,example) "16\#00E9"
+set AsciiOptions(4,info) "This notation is used in Postscript"
+set AsciiOptions(5,description)  "Generate hexadecimal numbers with prefix \#16r"
+set AsciiOptions(5,example) "\#16r00E9"
+set AsciiOptions(5,info) "This notation is used in Common Lisp"
+set AsciiOptions(6,description)  "Generate hexadecimal numbers with prefix 16\# and suffix \#"
+set AsciiOptions(6,example) "16\#00E9\#"
+set AsciiOptions(6,info) "This notation is used in Ada"
 
 set AsciiOptions(Example,info) "This column shows examples of the various types of notation.\nIn most cases the example is of the representation of the\nletter \u00E9 \"e with acute accent\". The exception is the\ncase in which the notation is different within the BMP and\noutside it. Here the first example is of \u00E9 and the second\nis of \"Linear B syllable E\"."
 set AsciiOptions(Description,info) "This column contains a description of the notation."
@@ -1060,10 +1095,19 @@ proc ConstructUACommandLine {ind} {
     if {$::AddSpaceP} {
 	lappend cl "-w"
     }
+    foreach s $::SubstitutionList {
+	lappend cl "-S"
+	lappend cl [format "%s:%s" [lindex $s 0] [lindex $s 1]]
+    }
+    foreach d $::DeletionList {
+	lappend cl "-S"
+	lappend cl [format "%s:" $d]
+    }
     set i [lindex $::OptionList $ind]
-    lappend cl [format "\-%s" [string toupper $i]]
+    lappend cl "-a"
+    lappend cl [format "%s" [string toupper $i]]
     if {[string equal $i "q"]} {
-	lappend cl "-H"
+	lappend cl "-a H"
     }
     return $cl
 }
@@ -1142,7 +1186,7 @@ proc ExecuteDaughter {ReallyP} {
 	    append msg [format " out of %d total characters" $TotalChars]
 	}
 	ShowMessage $msg
-	return $result;
+    	return $result;
     } else {
 	ShowMessage $cl;
     }
@@ -1243,7 +1287,7 @@ proc DescribeColumn {w c} {
 }
 
 puts "u2a $Version";
-puts "Copyright (C) 2005, 2006 William J. Poser.";
+puts "Copyright (C) 2005-2007 William J. Poser.";
 puts [_ "This program is free software; you can redistribute it
 and/or modify it under the terms of version 2 of the GNU General
 Public License as published by the Free Software Foundation."];
@@ -1304,15 +1348,18 @@ set GENOPTIONS .mf.f
 frame $GENOPTIONS -border 2 -relief ridge
 label $GENOPTIONS.tit -text "Miscellaneous Options"
 frame $GENOPTIONS.tbl
-checkbutton $GENOPTIONS.tbl.hu -variable HexUpperCaseP -text "Upper case hexadecimal numbers"
+checkbutton $GENOPTIONS.tbl.hu -variable HexUpperCaseP -text "Upper case hexadecimal numbers" \
+    -indicatoron 0
 checkbutton $GENOPTIONS.tbl.pt -variable ConvertAsciiP  -offvalue 0 -onvalue 1 \
-    -text "Convert ASCII characters"
+    -text "Convert ASCII characters"   -indicatoron 0
 checkbutton $GENOPTIONS.tbl.nt -variable PreserveNewlinesP  -offvalue 1 -onvalue 0 \
-    -text "Convert linefeed characters"
+    -text "Convert linefeed characters"     -indicatoron 0
 checkbutton $GENOPTIONS.tbl.wt -variable PreserveSpaceP  -offvalue 1 -onvalue 0 \
-    -text "Convert space characters"
-checkbutton $GENOPTIONS.tbl.at -variable AddSpaceP -text "Add a space after each item"
-checkbutton $GENOPTIONS.tbl.ipt -variable InputPureP -text "Input is pure"
+    -text "Convert space characters"     -indicatoron 0
+checkbutton $GENOPTIONS.tbl.at -variable AddSpaceP -text "Add a space after each item" \
+       -indicatoron 0
+checkbutton $GENOPTIONS.tbl.ipt -variable InputPureP -text "Input is pure" \
+    -indicatoron 0
 
 set msg "Here is where you can set various options. Note that some options
 apply only to uni2ascii and others only to ascii2uni. The options
