@@ -1,9 +1,9 @@
-/* Time-stamp: <2006-07-05 16:47:36 poser>
+/* Time-stamp: <2007-03-11 20:50:40 poser>
  *
  * Converts UTF-8 Unicode to pure 7-bit ASCII using any of a number
  * of different representations. 
  * 
- * Copyright (C) 2004-2006 William J. Poser (billposer@alum.mit.edu)
+ * Copyright (C) 2004-2007 William J. Poser (billposer@alum.mit.edu)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,6 +25,9 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
+#include <errno.h>
+#include <sys/types.h>
+#include <string.h>
 #ifdef HAVE_LOCALE_H
 #include <locale.h>
 #endif
@@ -39,6 +42,8 @@
 #include "endian.h"
 #include "utf8error.h"
 #include "exitcode.h"
+#include "formats.h"
+
 #if defined(__DATE__) && defined(__TIME__)
 #define HAVE_DATE_TIME
 char compdate[]= "Compiled " __DATE__ " " __TIME__ ;
@@ -50,6 +55,10 @@ char compdate[]= "";
 char version[]=PACKAGE_VERSION;
 char pgname[]="uni2ascii";
 
+#ifndef LOCALEDIR 
+#define LOCALEDIR "/usr/local/share/locale"
+#endif
+
 char msg [MSGSIZE];
 
 void
@@ -58,8 +67,12 @@ ShowVersion(void){
 #ifdef HAVE_DATE_TIME
   fprintf(stderr,"%s\n",compdate);
 #endif
-  fprintf(stderr,"Copyright (C) 2004-2006 William J. Poser\n");
-  fprintf(stderr,_("Released under the terms of the GNU General Public License.\n\n"));
+  fprintf(stderr,"Copyright (C) 2004-2007 William J. Poser\n");
+  fprintf(stderr,"This program is free software; you can redistribute\n\
+it and/or modify it under the terms of version 2 of\n\
+the GNU General Public License as published by the\n\
+Free Software Foundation.\n");
+  fprintf(stderr,"Report bugs to: billposer@alum.mit.edu.\n");
 }
 
 void
@@ -72,80 +85,58 @@ ideographic space, ethiopic word space and ogham space mark).\n\
 Options allow conversion of the ASCII characters as well\n\
 and for conversion even of newline and space characters.\n"));
   fprintf(stderr,_("Usage: %s [flags]\n"),pgname);
-  fprintf(stderr,_("       -c Convert circled and parenthesized characters.\n"));
-  fprintf(stderr,_("       -d Strip diacritics.\n"));
-  fprintf(stderr,_("       -e Convert to approximate ASCII equivalents.\n"));
-  fprintf(stderr,_("       -f Convert formal equivalents to ASCII.\n"));
   fprintf(stderr,_("       -h Print this usage message.\n"));
   fprintf(stderr,_("       -v Print version information.\n"));
-  fprintf(stderr,_("       -l Use lower-case a-f when generating hex.\n"));
-  fprintf(stderr,_("       -n Convert newlines.\n"));
-  fprintf(stderr,_("       -p Pure. Convert codepoints below 0x80 too.\n"));
+  fprintf(stderr,_("       -L List format specifications.\n"));
+  fprintf(stderr,_("       -E List the expansions performed by the -x option.\n"));
   fprintf(stderr,_("       -q Quiet. Do not chat unnecessarily while working.\n"));
-  fprintf(stderr,_("       -s Convert space characters.\n"));
-  fprintf(stderr,_("       -w Add a space after each converted item.\n"));
+  fprintf(stderr,_("       -a <format specification>.\n"));
+  fprintf(stderr,_("       -c Convert circled and parenthesized characters.\n"));
+  fprintf(stderr,_("       -e Convert to approximate ASCII equivalents.\n"));
+  fprintf(stderr,_("       -f Convert formal equivalents to ASCII.\n"));
   fprintf(stderr,_("       -x Expand certain single characters to equivalent sequences.\n"));
-  fprintf(stderr,_("          The expansions are:\n"));
-  fprintf(stderr,_("          U+00A9 COPYRIGHT SYMBOL            -> (c)\n"));
-  fprintf(stderr,_("          U+00DF SMALL LETTER SHARP S        -> ss\n"));
-  fprintf(stderr,_("          U+00C6 CAPITAL LETTER ASH          -> AE\n"));
-  fprintf(stderr,_("          U+00E6 SMALL LETTER ASH            -> ae\n"));
-  fprintf(stderr,_("          U+01F1 CAPITAL LETTER DZ           -> DZ\n"));
-  fprintf(stderr,_("          U+01F3 SMALLL LETTER DZ            -> dz\n"));
-  fprintf(stderr,_("          U+02A6 SMALLL LETTER TS DIGRAPH    -> ts\n"));
-  fprintf(stderr,_("          U+2026 HORIZONTAL ELLIPSIS         -> ...\n"));
-  fprintf(stderr,_("          U+22EF MIDLINE HORIZONTAL ELLIPSIS -> ...\n"));
-  fprintf(stderr,
-	  _("       -A Generate hexadecimal numbers with prefix U in angle-brackets(<U00E9>)\n"));
-  fprintf(stderr,
-	  _("       -B Generate backslash-x escaped hexadecimal numbers (\\x00E9)\n"));
-  fprintf(stderr,
-	  _("       -C Generate backslash-x escaped hexadecimal numbers in braces (\\x{00E9})\n"));
-  fprintf(stderr,
-	  _("       -D Generate decimal numeric character references (&#0233;)\n"));
-  fprintf(stderr,
-	  _("       -E Generate hexadecimal with prefix U (U00E9)\n"));
-  fprintf(stderr,
-	  _("       -F Generate hexadecimal with prefix u (u00E9)\n"));
-  fprintf(stderr,
-	  _("       -G Generate hexadecimal in single quotes with prefix X (X\'00E9\')\n"));
-  fprintf(stderr,
-	  _("       -H Generate hexadecimal numeric character references (&#x00E9;)\n"));
-  fprintf(stderr,
-	  _("       -I Generate hexadecimal UTF-8 with each byte's hex preceded by an =-sign (=C3=A9).\n\t\tThis is the Quoted Printable format defined by RFC 2045.\n"));
-  fprintf(stderr,
-	  _("       -J Generate hexadecimal UTF-8 with each byte's hex preceded by a %%-sign  (%%C3%%A9)\n\t\tThis is the URI escape format defined by RFC 2396.\n"));
-  fprintf(stderr,
-	  _("       -K Generate octal UTF-8 with backslash escapes (\303\251)\n"));
-  fprintf(stderr,
-	  _("       -L Generate \\U-escaped hex outside the BMP (U+0000-U+FFFF), \\u-escaped hex within.\n"));
-  fprintf(stderr,
-	  _("       -M Generate hexadecimal SGML numeric character references (\\#x00E9;)\n"));
-  fprintf(stderr,
-	  _("       -N Generate decimal SGML numeric character references (\\#0233;)\n"));
-  fprintf(stderr,
-	  _("       -O Generate octal escapes for the three low bytes in big-endian order (\\000\\000\\351)\n"));
-  fprintf(stderr,
-	  _("       -P Generate hexadecimal numbers with prefix U+ (U+00E9)\n"));
-  fprintf(stderr,
-	  _("       -Q Generate character entities where possible (&eacute;)\n"));
-  fprintf(stderr,
-	  _("       -R Generate raw hexadecimal numbers (00E9)\n"));
-  fprintf(stderr,
-	  _("       -S Generate hexadecimal escapes for the three low bytes in big-endian order (\\x00\\x00\\xE9)\n"));
-  fprintf(stderr,
-	  _("       -T Generate decimal escapes for the three low bytes in big-endian order (\\d000\\d000\\d233)\n"));
-  fprintf(stderr,
-	  _("       -U Generate \\u-escaped hex (\\u00E9)\n"));
-  fprintf(stderr,
-	  _("       -V Generate \\u-escaped decimal (\\u0233)\n"));
-  fprintf(stderr,
-	  _("       -X Generate standard form hexadecimal numbers (0x00E9)\n"));
+  fprintf(stderr,_("       -d Strip diacritics.\n"));
+  fprintf(stderr,_("       -S <Unicode:ASCII>  define a custom substitution\n"));
+  fprintf(stderr,_("       -n Convert newlines.\n"));
+  fprintf(stderr,_("       -s Convert space characters.\n"));
+  fprintf(stderr,_("       -p Convert codepoints below 0x80 too.\n"));
+  fprintf(stderr,_("       -l Use lower-case a-f when generating hex.\n"));
+  fprintf(stderr,_("       -w Add a space after each converted item.\n"));
   fprintf(stderr,
 	  _("       -Z <format> Use the specified format\n"));
   fprintf(stderr,_("Report bugs to: billposer@alum.mit.edu\n"));
 }
 
+void ListExpansions(void){
+  fprintf(stderr,_("The expansions performed by the -x flag are:\n"));
+  fprintf(stderr,_("          U+00A2 CENT SIGN                   -> cent\n"));
+  fprintf(stderr,_("          U+00A3 POUND SIGN                  -> pound\n"));
+  fprintf(stderr,_("          U+00A5 YEN SIGN                    -> yen\n"));
+  fprintf(stderr,_("          U+00A9 COPYRIGHT SYMBOL            -> (c)\n"));
+  fprintf(stderr,_("          U+00AE REGISTERED SYMBOL           -> (R)\n"));
+  fprintf(stderr,_("          U+00BC ONE QUARTER                 -> 1/4\n"));
+  fprintf(stderr,_("          U+00BD ONE HALF                    -> 1/2\n"));
+  fprintf(stderr,_("          U+00BE THREE QUARTERS              -> 3/4\n"));
+  fprintf(stderr,_("          U+00C6 CAPITAL LETTER ASH          -> AE\n"));
+  fprintf(stderr,_("          U+00DF SMALL LETTER SHARP S        -> ss\n"));
+  fprintf(stderr,_("          U+00E6 SMALL LETTER ASH            -> ae\n"));
+  fprintf(stderr,_("          U+0132 LIGATURE IJ                 -> IJ\n"));
+  fprintf(stderr,_("          U+0133 LIGATURE ij                 -> ij\n"));
+  fprintf(stderr,_("          U+0152 LIGATURE OE                 -> OE\n"));
+  fprintf(stderr,_("          U+0153 LIGATURE oe                 -> oe\n"));
+  fprintf(stderr,_("          U+01F1 CAPITAL LETTER DZ           -> DZ\n"));
+  fprintf(stderr,_("          U+01F2 MIXED LETTER Dz             -> Dz\n"));
+  fprintf(stderr,_("          U+01F3 SMALL LETTER dz             -> dz\n"));
+  fprintf(stderr,_("          U+02A6 SMALL LETTER TS DIGRAPH     -> ts\n"));
+  fprintf(stderr,_("          U+2026 HORIZONTAL ELLIPSIS         -> ...\n"));
+  fprintf(stderr,_("          U+20AC EURO SIGN                   -> euro\n"));
+  fprintf(stderr,_("          U+2190 LEFTWARDS ARROW             -> <-\n"));
+  fprintf(stderr,_("          U+2192 RIGHTWARDS ARROW            -> ->\n"));
+  fprintf(stderr,_("          U+21D0 LEFTWARDS DOUBLE ARROW      -> <=\n"));
+  fprintf(stderr,_("          U+21D2 RIGHTWARDS DOUBLE ARROW     -> =>\n"));
+  fprintf(stderr,_("          U+22EF MIDLINE HORIZONTAL ELLIPSIS -> ...\n"));
+}
+
 /* Swap the byte order of a 4 byte integer */
 
 void
@@ -172,29 +163,6 @@ lswab(unsigned long *val)
   *val = u.l;
 }
 
-#define HTMLX  0	/* HTML hex numeric */
-#define HTMLD  1
-#define SGMLX  2
-#define SGMLD  3
-#define BSLU   4
-#define BSLX   5   /* Tcl hex number as opposed to character code  */ 
-#define STDX   6
-#define RAWX   7
-#define BSLXB  8
-#define ABUX   9
-#define JUUX   10
-#define JuUX   11
-#define UPLX   12
-#define XQ     13
-#define BSLUD  14
-#define BYTEO  15
-#define BYTED  16
-#define BYTEH  17
-#define URLPX8 18
-#define URLEX8 19
-#define BSEO8  20
-
-
 static char *Formats [] = {
   "&#x%04x;",  /* HTMLX */
   "&#x%04X;",
@@ -202,7 +170,7 @@ static char *Formats [] = {
   "&#%04d;",
   "\\#x%04x;",  /* SGMLX */
   "\\#x%04X;",
-  "\\#%04d;",    /* HTMLD */
+  "\\#%04d;",    /* SGMLD */
   "\\#%04d;",
   "\\u%04x",   /* BSLU */
   "\\u%04X",   
@@ -210,6 +178,8 @@ static char *Formats [] = {
   "\\x%04X",
   "0x%04x",    /* STDX */
   "0x%04X",   
+  "#x%04x",    /* CLSX */
+  "#x%04X",   
   "%04x",       /* RAWX */
   "%04X",   
   "\\x{%04x}",	/* BSLXB */
@@ -226,12 +196,22 @@ static char *Formats [] = {
   "X\'%04X\'",
   "\\u%05d",	/* BSLUD */
   "\\u%05d",
+  "v%05d",	/* PERLV */
+  "v%05d",
+  "$%04x",  	/* DOLLAR */
+  "$%04X",
+  "16#%04x",	/* PSPT */
+  "16#%04X",
+  "#16r%04x",	/* CLR */
+  "#16r%04X",
+  "16#%04x#",	/* ADA */
+  "16#%04X#",
   "\\%03hho\\%03hho\\%03hho",	/* BYTEO */
   "\\%03hho\\%03hho\\%03hho",
   "\\d%03hhd\\d%03hhd\\d%03hhd",	/* BYTED */
   "\\d%03hhd\\d%03hhd\\d%03hhd",
   "\\x%02hhx\\x%02hhx\\x%02hhx",/* BYTEH */
-  "\\x%02hhX\\x%02hhX\\x%02hhX"	
+  "\\x%02hhX\\x%02hhX\\x%02hhX"
   };
 
 /*  Returns 1 if it handled the character, 0 if not. */
@@ -433,22 +413,18 @@ int AscifyEnclosed(UTF32 c) {
 int AscifyEquiv(UTF32 c) {
   switch (c)
     {
-    case 0x0026:		/* ampersand */
     case 0xFE60:
-      putchar(0x26);
+      putchar(0x26);		/* ampersand */
       break;
-    case 0x002A:		/* asterisk */
     case 0x204E:
     case 0x2217:
     case 0x2731:
     case 0xFE61:
-      putchar(0x2A);
+      putchar(0x2A);		/* asterisk */
       break;
-    case 0x002B:		/* plus sign */
-    case 0xFE62:
+    case 0xFE62:		/* plus sign */
       putchar(0x2B);
       break;
-    case 0x002D:		/* Hyphen */
     case 0x00AD:
     case 0x2010:
     case 0x2011:
@@ -456,9 +432,14 @@ int AscifyEquiv(UTF32 c) {
     case 0x2013:
     case 0x2014:
     case 0x2212:
-      putchar(0x2D);
+    case 0x2500:
+    case 0x2501:
+      putchar(0x2D);		/* Hyphen */
       break;
-    case 0x00AB:		/* Double quote */
+    case 0x2502:
+    case 0x2503:
+      putchar(0x7C);		/* Vertical line */
+    case 0x00AB:
     case 0x00BB:
     case 0x201C:
     case 0x201D:
@@ -466,19 +447,17 @@ int AscifyEquiv(UTF32 c) {
     case 0x201F:
     case 0x275D:
     case 0x275E:
-      putchar(0x22);
+      putchar(0x22);		/* Double quote */
       break;
-    case 0x0060:		/* left single quote */
     case 0x2018:
     case 0x201A:
     case 0x201B:
     case 0x2039:
-      putchar(0x60);
+      putchar(0x60);		/* left single quote */
       break;
-    case 0x0027:		/* right or neutral single quote */
     case 0x2019:
     case 0x203A:
-      putchar(0x27);
+      putchar(0x27);		/* right or neutral single quote */
       break;
     case 0x222A:		/* Union */
       putchar('U');
@@ -511,6 +490,12 @@ int AscifyEquiv(UTF32 c) {
 int AscifyDiacritics(UTF32 c) {
   switch (c)
     {
+    case 0x226E:
+      putchar(0x3C);
+      break;
+    case 0x226F:
+      putchar(0x3E);
+      break;
     case 0x00C0:
     case 0x00C1:
     case 0x00C2:
@@ -620,6 +605,11 @@ int AscifyDiacritics(UTF32 c) {
     case 0x0124:
     case 0x0126:
     case 0x021E:
+    case 0x1E22:
+    case 0x1E24:
+    case 0x1E26:
+    case 0x1E28:
+    case 0x1E2A:
       putchar(0x48);		/* H */
       break;
     case 0x00CC:
@@ -646,6 +636,7 @@ int AscifyDiacritics(UTF32 c) {
     case 0x0134:
       putchar(0x4A);		/* J */
       break;
+    case 0x0136:
     case 0x0198:
     case 0x01E8:
     case 0x1E30:
@@ -653,6 +644,9 @@ int AscifyDiacritics(UTF32 c) {
     case 0x1E34:
       putchar(0x4B);		/* K */
       break;
+    case 0x0139:
+    case 0x013B:
+    case 0x013D:
     case 0x023D:
     case 0x1E36:
     case 0x1E38:
@@ -666,6 +660,9 @@ int AscifyDiacritics(UTF32 c) {
       putchar(0x4D);		/* M */
       break;
     case 0x00D1:
+    case 0x0143:
+    case 0x0145:
+    case 0x0147:
     case 0x019D:
     case 0x01F8:
     case 0x0220:
@@ -720,6 +717,9 @@ int AscifyDiacritics(UTF32 c) {
     case 0x1E56:
       putchar(0x50);		/* P */
       break;
+    case 0x0154:
+    case 0x0156:
+    case 0x0158:
     case 0x0210:
     case 0x0212:
     case 0x1E58:
@@ -728,6 +728,10 @@ int AscifyDiacritics(UTF32 c) {
     case 0x1E5E:
       putchar(0x52);		/* R */
       break;
+    case 0x015A:
+    case 0x015C:
+    case 0x015E:
+    case 0x0160:
     case 0x0218:
     case 0x1E60:
     case 0x1E62:
@@ -736,6 +740,8 @@ int AscifyDiacritics(UTF32 c) {
     case 0x1E68:
       putchar(0x53);		/* S */
       break;
+    case 0x0162:
+    case 0x0164:
     case 0x01AC:
     case 0x01AE:
     case 0x021A:
@@ -965,6 +971,7 @@ int AscifyDiacritics(UTF32 c) {
     case 0x029D:
       putchar(0x6A);		/* j */
       break;
+    case 0x0137:
     case 0x0199:
     case 0x01E9:
     case 0x1D84:
@@ -973,6 +980,9 @@ int AscifyDiacritics(UTF32 c) {
     case 0x1E35:
       putchar(0x6B);		/* k */
       break;
+    case 0x013A:
+    case 0x013C:
+    case 0x013E:
     case 0x019A:
     case 0x0234:
     case 0x026B:
@@ -994,6 +1004,9 @@ int AscifyDiacritics(UTF32 c) {
       putchar(0x6D);		/* m */
       break;
     case 0x00F1:
+    case 0x0144:
+    case 0x0146:
+    case 0x0148:
     case 0x019E:
     case 0x01F9:
     case 0x0272:
@@ -1012,6 +1025,9 @@ int AscifyDiacritics(UTF32 c) {
     case 0x00F5:
     case 0x00F6:
     case 0x00F8:
+    case 0x014D:
+    case 0x014F:
+    case 0x0151:
     case 0x01A1:
     case 0x01D2:
     case 0x01EB:
@@ -1052,6 +1068,9 @@ int AscifyDiacritics(UTF32 c) {
     case 0x02A0:
       putchar(0x71);		/* q */
       break;
+    case 0x0155:
+    case 0x0157:
+    case 0x0159:
     case 0x0211:
     case 0x0213:
     case 0x027C:
@@ -1066,9 +1085,12 @@ int AscifyDiacritics(UTF32 c) {
     case 0x1E5F:
       putchar(0x72);		/* r */
       break;
+    case 0x015B:
+    case 0x015D:
     case 0x0219:
     case 0x023F:
     case 0x0282:
+    case 0x0161:
     case 0x1D74:
     case 0x1D8A:
     case 0x1E61:
@@ -1078,6 +1100,7 @@ int AscifyDiacritics(UTF32 c) {
     case 0x1E69:
       putchar(0x73);		/* s */
       break;
+    case 0x0165:
     case 0x01AB:
     case 0x01AD:
     case 0x021B:
@@ -1094,6 +1117,9 @@ int AscifyDiacritics(UTF32 c) {
     case 0x00FA:
     case 0x00FB:
     case 0x00FC:
+    case 0x0169:
+    case 0x016D:
+    case 0x016F:
     case 0x0171:
     case 0x0173:
     case 0x01B0:
@@ -1272,7 +1298,6 @@ int AscifyDiacritics(UTF32 c) {
   return 1;
 } /* End of AscifyDiacritics */
 
-
 /* 
  * Convert stylistic variants to plain equivalent.
  *
@@ -2318,11 +2343,49 @@ int AscifyStyle(UTF32 c) {
 int ExpandToAscii(UTF32 c) {
   switch(c)
     {
+    case 0x00A2:	/* Cent sign */
+      putchar('c');
+      putchar('e');
+      putchar('n');
+      putchar('t');
+      break;
+    case 0x00A3:	/* Pound sign */
+      putchar('p');
+      putchar('o');
+      putchar('u');
+      putchar('n');
+      putchar('d');
+      break;
+    case 0x00A5:	/* Yen sign */
+      putchar('y');
+      putchar('e');
+      putchar('n');
+      break;
     case 0x00A9:	/* Copyright symbol */
       putchar('(');
       putchar('c');
       putchar(')');
       break;
+    case 0x00AE:	/* Registered symbol */
+      putchar('(');
+      putchar('R');
+      putchar(')');
+      break;
+    case 0x00BC:	/* One quarter */
+      putchar('1');
+      putchar('/');
+      putchar('4');
+      break;
+    case 0x00BD:	/* One half */
+      putchar('1');
+      putchar('/');
+      putchar('2');
+      break;
+    case 0x00BE:	/* Three quarters */
+      putchar('3');
+      putchar('/');
+      putchar('4');
+      break;
     case 0x00C6:	/* Ash */
       putchar('A');
       putchar('E');
@@ -2335,10 +2398,30 @@ int ExpandToAscii(UTF32 c) {
       putchar('s');
       putchar('s');
       break;
+    case 0x0132:	/* Ligature IJ */
+      putchar('I');
+      putchar('J');
+      break;
+    case 0x0133:	/* Ligature ij */
+      putchar('i');
+      putchar('j');
+      break;
+    case 0x0152:	/* Ligature OE */
+      putchar('O');
+      putchar('E');
+      break;
+    case 0x0153:	/* Ligature oe */
+      putchar('o');
+      putchar('e');
+      break;
     case 0x01F1:	/* Capital DZ */
       putchar('D');
       putchar('Z');
       break;
+    case 0x01F2:	/* Mixed Dz */
+      putchar('D');
+      putchar('z');
+      break;
     case 0x01F3:	/* Small DZ */
       putchar('d');
       putchar('z');
@@ -2347,6 +2430,12 @@ int ExpandToAscii(UTF32 c) {
       putchar('t');
       putchar('s');
       break;
+    case 0x20AC:	/* Euro sign */
+      putchar('e');
+      putchar('u');
+      putchar('r');
+      putchar('o');
+      break;
     case 0x2026:	/* Horizontal ellipsis */
     case 0x22EF:	/* Midline horizontal ellipsis */
       putchar('.');
@@ -2375,6 +2464,89 @@ int ExpandToAscii(UTF32 c) {
   return 1;
 }
 
+void
+CheckNumberError(char *pgname,char *s, char c){
+  if(errno == ERANGE){
+    fprintf(stderr,"%s: %s falls outside of the representable range\n",pgname,s);
+    exit(OTHERERROR);
+  }
+  if(errno == EINVAL){
+    fprintf(stderr,"%s: %s is ill-formed\n",pgname,s);
+    exit(OTHERERROR);
+  }
+  if(c != '\0'){
+    fprintf(stderr,"%s: %s is ill-formed\n",pgname,s);
+    exit(OTHERERROR);
+  }
+}
+
+struct subpair {
+  UTF32 u;
+  char  a;
+};
+
+struct subpair *SubList = NULL;
+int SubCnt = 0;
+int SubsAvailable = 0;
+
+/*
+ * If people start using lots of custom substitutions, a more efficient
+ * search may be advisable. For the time being we just do a linear search.
+ */
+
+SubstituteChar(UTF32 c) {
+  int i;
+  for(i = 0; i < SubCnt; i++) {
+    if(c == SubList[i].u) {
+      if(SubList[i].a) {
+	putchar(SubList[i].a);
+	return 1;
+      }
+      else return (-1);		/* Signal deletion */
+    }
+  }
+  return 0;
+}
+
+AddCustomSubstitution(char *str){
+  char *Left;
+  char *Right;
+  char *Delim;
+  char *endptr;
+  unsigned long UCode;
+  unsigned long ACode;
+
+  Delim = strchr(str,':');
+  if(Delim == NULL) {
+    fprintf(stderr,"Substition specification %s is ill-formed.\n",str);
+    exit(BADOPTIONARG);
+  }
+  *Delim = '\0';
+  Left = str;
+  Right = Delim+1;
+  UCode = strtoul(Left,&endptr,0);
+  CheckNumberError(pgname,Left,*endptr);
+  if(*Right == '\0') ACode = 0;
+  else {
+    ACode = strtoul(Right,&endptr,0);
+    CheckNumberError(pgname,Right,*endptr);
+  }
+  if(ACode > 127L) {
+    fprintf(stderr,"Ascii code for substitution must be between 1 and 127 inclusive.\n");
+    exit(BADOPTIONARG);
+  }
+  SubCnt++;
+  if(SubCnt > SubsAvailable) {
+    SubList = realloc(SubList,sizeof(struct subpair) * (SubsAvailable+=4));
+    if(SubList == NULL) {
+      fprintf(stderr,"Unable to allocate storage for substitution list.\n");
+      exit(OUTOFMEMORY);
+    }
+  }
+  SubList[SubCnt-1].u = (UTF32) UCode;
+  SubList[SubCnt-1].a = (char)  ACode;
+}
+
 int main (int ac, char *av[])
 {
 
@@ -2385,22 +2557,29 @@ int main (int ac, char *av[])
   int ucnt;			/* Index into current UTF* string  */
   int UCBytes;
   int infd;
-  int FType;
+  int FType = STDX;
+  short UseEntitiesP = 0;
   int UTF8Type = 0;
+  short BMPSplitP = 0;
+  char *fmt = NULL;
+  short dummy;			/* For compatibility with ascii2uni */
+
+
   unsigned char b1,b2,b3; /* The low three bytes of a 4 byte UTF-32 character */
 
   unsigned long ByteCnt;
   unsigned long CharCnt;
   unsigned long ConvertedCnt;
+  unsigned long SubstitutedCnt;
+  unsigned long DeletedCnt;
 
   short PureP = 0;
   short PreserveNewlinesP = 1;
   short PreserveSpacesP = 1;
   short VerboseP = 1;
-  short BMPSplitP =0;
   short HexUpperP = 1;	/* Use X or x in  hex ?  */
   short AddWhitespaceP = 0;
-  short UseEntitiesP = 0;
+
   short AscifyStyleP = 0;
   short AscifyDiacriticsP = 0;
   short AscifyEnclosedP = 0;
@@ -2410,20 +2589,24 @@ int main (int ac, char *av[])
 
   char *AboveBMPfmt;
   char *WithinBMPfmt;
-  char *fmt = NULL;
   char *e;
   unsigned char *ustr;	/* UTF8 string */
 
   extern int optind;
   extern int opterr;
   extern int optopt;
-  extern UTF32 Get_UTF32_From_UTF8 (int,int *,unsigned char **);
+  extern UTF32 UTF8in(int,int *,unsigned char **);
+
+  extern int CountSlots(char *);
+  extern void ListFormatArguments(short);
+  extern void SetFormat(char *, int *, short *,int *, short *, short *);
 
   opterr = 0;
   ByteCnt = 0L;
   CharCnt = 0L;
   ConvertedCnt = 0L;
-  FType = STDX;			/* Default output is standard hex */
+  SubstitutedCnt = 0L;
+  DeletedCnt = 0L;
 
 #ifdef HAVE_SETLOCALE
    setlocale(LC_ALL,"");
@@ -2435,16 +2618,14 @@ int main (int ac, char *av[])
 
   /* Handle command line arguments */
 
-  while( (oc = getopt(ac,av,"cdefhlnpqsvwxABCDEFGHIJKLMNOPQRSTUVXVZ:")) != EOF){
+  while( (oc = getopt(ac,av,":a:cdeEfhlLnpqsS:vwxZ:")) != EOF){
     switch(oc){
-    case 'O':
-      FType = BYTEO;
-      break;
-    case 'S':
-      FType = BYTEH;
-      break;
-    case 'T':
-      FType = BYTED;
+    case 'a':
+      SetFormat(optarg,&FType,&UseEntitiesP, &UTF8Type, &BMPSplitP,&dummy);
+       if(FType == FMT_UNKNOWN) {
+	 fprintf(stderr,"Format specification %s not recognized.\n",optarg);
+	 exit(BADOPTIONARG);
+       }
       break;
     case 'c':
       AscifyEnclosedP = 1;
@@ -2452,6 +2633,9 @@ int main (int ac, char *av[])
     case 'd':
       AscifyDiacriticsP = 1;
       break;
+    case 'E':
+      ListExpansions();
+      exit(INFO);
     case 'e':
       AscifyEquivP = 1;
       break;
@@ -2461,6 +2645,9 @@ int main (int ac, char *av[])
     case 'l':
       HexUpperP = 0;
       break;
+    case 'L':
+      ListFormatArguments(1);
+      exit(INFO);
     case 'n':
       PreserveNewlinesP = 0;
       break;
@@ -2470,77 +2657,21 @@ int main (int ac, char *av[])
     case 's':
       PreserveSpacesP = 0;
       break;
+    case 'S':
+      AddCustomSubstitution(optarg);
+      break;
     case 'x':
       ExpandP = 1;
       break;
     case 'w':
       AddWhitespaceP = 1;
       break;
-    case 'A':
-      FType = ABUX;
-      break;
-    case 'B':
-      FType = BSLX;
-      break;
-    case 'C':
-      FType = BSLXB;
-      break;
-    case 'D':
-      FType = HTMLD;
-      break;
-    case 'E':
-      FType = JUUX;
-      break;
-    case 'F':
-      FType = JuUX;
-      break;
-    case 'G':
-      FType = XQ;
-      break;
-    case 'H':
-      FType = HTMLX;
-      break;
-    case 'I':
-      UTF8Type = 2;
-      break;
-    case 'J':
-      UTF8Type = 1;
-      break;
-    case 'K':
-      UTF8Type = 3;
-      break;
-    case 'L':
-      FType = BSLU;
-      BMPSplitP = 1;
-      break;
-    case 'M':
-      FType = SGMLX;
-      break;
-    case 'N':
-      FType = SGMLD;
-      break;
-    case 'P':
-      FType = UPLX;
-      break;
-    case 'Q':
-      UseEntitiesP = 1;
-      break;
-    case 'R':
-      FType = RAWX;
-      break;
-    case 'U':
-      FType = BSLU;
-      BMPSplitP = 0;
-      break;
-    case 'V':
-      FType = BSLUD;
-      BMPSplitP = 0;
-      break;
-    case 'X':
-      FType = STDX;
-      break;
     case 'Z':
       fmt = optarg;
+      if(CountSlots(fmt) > 1) {
+	fprintf(stderr,"You may not supply a format with more than one empty slot.\n");
+	exit(BADOPTIONARG);
+      }
       break;
     case 'h':
       ShowUsage();
@@ -2565,6 +2696,10 @@ int main (int ac, char *av[])
   } 
   infd = fileno(stdin);
 
+  if(VerboseP) {
+    if(SubCnt) fprintf(stderr,"Custom substitutions defined: %d\n",SubCnt);
+  }
+
   /* Combinatorial checks */
 
   if(UseEntitiesP && !(FType == HTMLX || FType == HTMLD)) {
@@ -2602,23 +2737,49 @@ int main (int ac, char *av[])
 
   if(UTF8Type) {
     if(AscifyStyleP|AscifyDiacriticsP|AscifyEquivP|AscifyEnclosedP|ExpandP) {
-      while ( (c = Get_UTF32_From_UTF8(infd,&UCBytes,&ustr)) <= UNI_MAX_UTF32) {
+      while ( (c = UTF8in(infd,&UCBytes,&ustr)) <= UNI_MAX_UTF32) {
 	ByteCnt+=UCBytes;
 	CharCnt++;
+	if(SubCnt) {
+	  switch(SubstituteChar(c))
+	    {
+	    case 1:
+	      SubstitutedCnt++;
+	      continue;
+	    case -1:
+	      DeletedCnt++;
+	      continue;
+	    }
+	}
 	if(AscifyStyleP) {
-	  if(AscifyStyle(c)) continue; 
+	  if(AscifyStyle(c))  {
+	    SubstitutedCnt++;
+	    continue;
+	  }
 	} 
 	if(AscifyDiacriticsP) {
-	  if(AscifyDiacritics(c)) continue; 
+	  if(AscifyDiacritics(c))  {
+	    SubstitutedCnt++;
+	    continue;
+	  }
 	} 
 	if(ExpandP) {
-	  if(ExpandToAscii(c)) continue;
+	  if(ExpandToAscii(c)) {
+	    SubstitutedCnt++;
+	    continue;
+	  }
 	}
 	if(AscifyEquivP) {
-	  if(AscifyEquiv(c)) continue; 
+	  if(AscifyEquiv(c)) {
+	    SubstitutedCnt++;
+	    continue;
+	  }
 	} 
 	if(AscifyEnclosedP) {
-	  if(AscifyEnclosed(c)) continue; 
+	  if(AscifyEnclosed(c)) {
+	    SubstitutedCnt++;
+	    continue;
+	  }
 	} 
 	ucnt= 0;
 	while(ucnt < UCBytes) {
@@ -2666,24 +2827,50 @@ int main (int ac, char *av[])
     }
   } /* End of UTF8 types */
 
-  while ( (c = Get_UTF32_From_UTF8(infd,&UCBytes,&ustr)) <= UNI_MAX_UTF32){
+  while ( (c = UTF8in(infd,&UCBytes,&ustr)) <= UNI_MAX_UTF32){
     ByteCnt+=UCBytes;
     CharCnt++;
+    if(SubCnt) {
+      switch(SubstituteChar(c))
+	{
+	case 1:
+	  SubstitutedCnt++;
+	  continue;
+	case -1:
+	  DeletedCnt++;
+	  continue;
+	}
+    }
     if(AscifyStyleP) {
-      if(AscifyStyle(c)) continue; 
+      if(AscifyStyle(c))  {
+	SubstitutedCnt++;
+	continue;
+      }
     } 
     if(AscifyDiacriticsP) {
-      if(AscifyDiacritics(c)) continue; 
+      if(AscifyDiacritics(c))  {
+	SubstitutedCnt++;
+	continue;
+      }
     } 
     if(ExpandP) {
-      if(ExpandToAscii(c)) continue;
+      if(ExpandToAscii(c)) {
+	SubstitutedCnt++;
+	continue;
+      }
     }
     if(AscifyEquivP) {
-      if(AscifyEquiv(c)) continue; 
+      if(AscifyEquiv(c)) {
+	SubstitutedCnt++;
+	continue;
+      }
     } 
     if(AscifyEnclosedP) {
-      if(AscifyEnclosed(c)) continue; 
-    }
+      if(AscifyEnclosed(c)) {
+	SubstitutedCnt++;
+	continue;
+      }
+    } 
     switch (c) {
     case 0x000A:
       if(PreserveNewlinesP) putchar((int)c);
@@ -2771,7 +2958,18 @@ int main (int ac, char *av[])
     break;
   }
 
+#ifdef NEWSUMMARY
+  if(VerboseP) {
+    fprintf(stderr,_("Total input characters          %12ld\n"),CharCnt);
+    fprintf(stderr,_("Characters converted to escapes %12ld\n"),ConvertedCnt);
+    fprintf(stderr,_("Characters replaced with ASCII  %12ld\n"),SubstitutedCnt);
+    fprintf(stderr,_("Characters deleted              %12ld\n"),DeletedCnt);
+  }
+#else  
   if(VerboseP) fprintf(stderr,_("%1$ld tokens converted out of %2$ld characters\n"),ConvertedCnt,CharCnt);
+#endif
   exit(SUCCESS);
 }
 
+
+
